@@ -4,7 +4,7 @@ SCRATCH=scratch/$(basename $0 .sh)
 rm -rf ${SCRATCH}
 mkdir -p ${SCRATCH}
 
-cp ${srcdir}/no-rpath ${SCRATCH}/
+cp no-rpath ${SCRATCH}/
 
 oldRPath=$(../src/patchelf --print-rpath ${SCRATCH}/no-rpath)
 if test -n "$oldRPath"; then exit 1; fi
@@ -18,6 +18,4 @@ if ! echo "$newRPath" | grep -q '/foo:/bar'; then
     exit 1
 fi
 
-if [ "$(uname -m)" = i686 -a "$(uname -s)" = Linux ]; then
-    cd ${SCRATCH} && ./no-rpath
-fi
+cd ${SCRATCH} && ./no-rpath
